package nxm.sys.lib;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;

import nxm.sys.inc.InternalUseOnly;

/**
 * A class that will parse and maintain all the header information from a file.
 * @author Kevin Esterheld
 * @since NeXtMidas 3.9.2
 */
@InternalUseOnly
public class DataFileHeader {

  private DataFile df;

  private byte[] hbuf;

  // header values
  private Midas M;
  private Keywords extendedHeaderKeywords;
  private Keywords mainKeywords;
  private String headerRep;
  private String dataRep;
  private int type;
  private String format;
  private byte formatMode;
  private byte formatType;
  private double xStart;
  private double xDelta;
  private double yStart;
  private double yDelta;
  private short dataMask;
  private double dataSize;
  private double dataStart;
  private int extSize;
  private int extStart;
  private String version;
  private int pipeId;
  private int detached;
  private int protectedFromOverwrite;
  private double timeCode;
  private int keyLength;
  private short inlet;
  private short outlet;
  private int pipeLoc;
  private int pipeSize;
  private double inByte;
  private double outByte;
  InputStream fis;
  IOResource res;
  PipeResource resPipe;
  private boolean ableToParseHeader = false;

  /** DataFileHeader Constructor
      @param df DataFile to be referenced
   */
  public DataFileHeader (DataFile df) {
    this.df = df;
    this.M = df.M;
    readHeader();
  }

  /**
   * @return the keywords
   */
  @InternalUseOnly
  public Keywords getExtendedHeaderKeywords () {
    if(res instanceof PipeResource) {
      return resPipe.pipe.getDataFile().getKeywordsObject();
    }
    if (extendedHeaderKeywords == null && ableToParseHeader) {
      long off = ((long) extStart) * DataFile.HEADER_SIZE;
      if (!res.isOpen) res.open();
      // extended keywords
      // get the length of the keywords
      int len = (extSize + 511) / DataFile.HEADER_SIZE * DataFile.HEADER_SIZE;
      // create keywords buffer
      byte[] bufWithKeys = new byte[len];
      // read the keywords into the buffer
      int bytes = res.read(bufWithKeys, 0, extSize, off);
      // create and init the keywords with bufwithkeys
      extendedHeaderKeywords = new Keywords();
      extendedHeaderKeywords.init(bufWithKeys, bytes, hbuf[DataFile.OFF_HEADREP]);
      res.close();
    }
    return extendedHeaderKeywords;
  }

  /**
   * Method to grab only the header information from a data file. This is helpful when it is necessary to get the header
   * information often from the file.
   */
  private void readHeader () {
    try {


      res = M.io.findResource(df.getFileName(), -1, "", null, false);

      hbuf = new byte[DataFile.HEADER_SIZE];

      if(res instanceof PipeResource) {
//        if(true) return;
        resPipe = (PipeResource) res;
        res.exists();
        resPipe.df = df;
        Convert.packD(hbuf, DataFile.OFF_DATASTART,0);
        Convert.packL(hbuf, DataFile.OFF_PIPE,0);
        resPipe.pipeSize = 131072;
        resPipe.open();
        resPipe.readHeader(hbuf);
      }
      else {

        res.open();
        fis = res.getInputStream();
        fis.read(hbuf);
      }

      version = Convert.unpackS(hbuf, DataFile.OFF_VERSION, 4);
      headerRep = Convert.unpackS(hbuf, DataFile.OFF_HEADREP, 4);
      dataRep = Convert.unpackS(hbuf, DataFile.OFF_DATAREP, 4);
      detached = Convert.unpackL(hbuf, DataFile.OFF_DETACHED);
      protectedFromOverwrite = Convert.unpackL(hbuf, DataFile.OFF_PROTECTED);
      pipeId = Convert.unpackL(hbuf, DataFile.OFF_PIPE);
      extStart = Convert.unpackL(hbuf, DataFile.OFF_EXTSTART);
      extSize = Convert.unpackL(hbuf, DataFile.OFF_EXTSIZE);
      dataStart = Convert.unpackD(hbuf, DataFile.OFF_DATASTART);
      dataSize = Convert.unpackD(hbuf, DataFile.OFF_DATASIZE);
      type = Convert.unpackL(hbuf, DataFile.OFF_TYPE);
      format = Convert.unpackS(hbuf, DataFile.OFF_FORMAT, 4);
      formatMode = Convert.unpackB(hbuf, DataFile.OFF_FORMAT);
      formatType = Convert.unpackB(hbuf, DataFile.OFF_FORMAT + 1);
      dataMask = Convert.unpackI(hbuf, DataFile.OFF_FLAGMASK);
      timeCode = Convert.unpackD(hbuf, DataFile.OFF_TIMECODE);
      inlet = Convert.unpackI(hbuf, DataFile.OFF_INLET);
      outlet = Convert.unpackI(hbuf, DataFile.OFF_OUTLETS);
      pipeLoc = Convert.unpackL(hbuf, DataFile.OFF_PIPELOC);
      pipeSize = Convert.unpackL(hbuf, DataFile.OFF_PIPESIZE);
      inByte = Convert.unpackD(hbuf, DataFile.OFF_INBYTE);
      outByte = Convert.unpackD(hbuf, DataFile.OFF_OUTBYTE);
      // parse outbytes? - 96 - 159
      keyLength = Convert.unpackL(hbuf, DataFile.OFF_KEYLGTH);
      xStart = Convert.unpackD(hbuf, DataFile.OFF_ADJUNCT + DataFile.ADJUNCT_XSTART_OFF);
      xDelta = Convert.unpackD(hbuf, DataFile.OFF_ADJUNCT + DataFile.ADJUNCT_XDELTA_OFF);
      if (type / 1000 >= 2) {
        yStart = Convert.unpackD(hbuf, DataFile.OFF_ADJUNCT + DataFile.ADJUNCT_YSTART_OFF);
        yDelta = Convert.unpackD(hbuf, DataFile.OFF_ADJUNCT + DataFile.ADJUNCT_YDELTA_OFF);
      }
      // The following is keywords magic.
      ableToParseHeader = true;
      res.close();
      if (fis!=null) fis.close();
    } catch (Exception e) {
      Shell.getMidasContext().warning("Unable to parse the header from " + df.getFileName());
    }
  }

  /**
   * Add add a header keyword to the file.
   * This will allow the adding of custom types and values.
   * @param type The char type to identify the value type
   * @param val the object to be added.
   * @param name The key name of the keyword.
   * @since 3.9.3
   */
  @InternalUseOnly
  public void addToFile(char type, Object val, String name) {
    FileInputStream fis;
    byte[] buf;
    File file = new File(df.filename);

    try {
      fis = new FileInputStream(file);
      buf = new byte[(int) file.length()];
      fis.read(buf);
      fis.close();
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
      return;
    }
    byte[] data = packTestDataToBuffer(type,val);
    int off = ((extStart*512) + extSize);
    if(off==0) off = (int) (dataStart+dataSize);
    byte ltag = (byte)name.length();
    int len = data.length;
    int length = (len+ltag+15)/8*8;
    short lext = (short)(length - len);
    if (buf.length < (length+off)) {
      int size = buf.length;
      byte[] bufo = buf;
      buf = new byte[length+off];
      System.arraycopy(bufo, 0, buf, 0, size);
    }
    Convert.packL(buf, off+0, length);
    Convert.packI(buf, off+4, lext);
    buf[off+6] = ltag;
    buf[off+7] = (byte) type;
    System.arraycopy (data,0, buf,off+Keywords.HEADER_LEN, len);
    for (int i=0,j=off+Keywords.HEADER_LEN+len; i<ltag;) buf[j++] = (byte)name.charAt(i++); // set tag name
    Convert.packL(buf, DataFile.OFF_EXTSIZE, extSize+length);
    Convert.packL(buf, DataFile.OFF_EXTSTART, off/512);

    // update changed fields
    extSize += length;
    try (FileOutputStream fos = new FileOutputStream(file); BufferedOutputStream bos = new BufferedOutputStream(fos);){
      bos.write(buf);
    } catch (IOException e) {
      e.printStackTrace();
    }
    readHeader();
  }

  /** pack test data into buffer - for testing only
   *
   *  @param type type of data 'U' or 'V'
   *  @param val  value
   *  @return byte buffer
   */
  @InternalUseOnly
  public byte[] packTestDataToBuffer(char type, Object val) {
    byte[] res = new byte[type=='U' ? 2 : 4];
    if(type=='U') {
      Convert.packI(res, 0,(short) val);
    }
    else {
      Convert.packL(res, 0,(int) val);
    }

    return res;
  }

  /**
   * @return the headerbuf
   */
  @InternalUseOnly
  public byte[] getHeaderbuf () {
    return hbuf;
  }

  /**
   * @return the mainKeywords
   */
  @InternalUseOnly
  public Keywords getMainKeywords () {
    if(res instanceof PipeResource) return ((PipeResource)res).pipe.getDataFile().getKeywordsObject();
    if (mainKeywords == null && ableToParseHeader) {
      if (!res.isOpen) res.open();
      byte[] mainKeysBuf = new byte[keyLength];
      res.read(mainKeysBuf, 0, keyLength, DataFile.OFF_KEYWORDS);
      String keytemp = new String(mainKeysBuf);
      mainKeywords = new Keywords();
      mainKeywords.init(null, 0, hbuf[DataFile.OFF_HEADREP]);
      String[] keywordsRaw = keytemp.split("=");
      for (int i = 0; i < keywordsRaw.length - 1; i += 2) {
        mainKeywords.put(keywordsRaw[i], keywordsRaw[i + 1]);
      }
    }
    return mainKeywords;
  }

  /**
   * @return the headerRep
   */
  @InternalUseOnly
  public String getHeaderRep () {
    return headerRep;
  }

  /**
   * @return the dataRep
   */
  @InternalUseOnly
  public String getDataRep () {
    return dataRep;
  }

  /**
   * @return the type
   */
  @InternalUseOnly
  public int getType () {
    return type;
  }

  /**
   * @return the format
   */
  @InternalUseOnly
  public String getFormat () {
    return format;
  }

  /**
   * @return the formatMode
   */
  @InternalUseOnly
  public byte getFormatMode () {
    return formatMode;
  }

  /**
   * @return the formatType
   */
  @InternalUseOnly
  public byte getFormatType () {
    return formatType;
  }

  /**
   * @return the xStart
   */
  @InternalUseOnly
  public double getxStart () {
    return xStart;
  }

  /**
   * @return the xDelta
   */
  @InternalUseOnly
  public double getxDelta () {
    return xDelta;
  }

  /**
   * @return the yStart
   */
  @InternalUseOnly
  public double getyStart () {
    return yStart;
  }

  /**
   * @return the yDelta
   */
  @InternalUseOnly
  public double getyDelta () {
    return yDelta;
  }

  /**
   * @return the dataMask
   */
  @InternalUseOnly
  public short getDataMask () {
    return dataMask;
  }

  /**
   * @return the dataSize
   */
  @InternalUseOnly
  public double getDataSize () {
    return dataSize;
  }

  /**
   * @return the dataStart
   */
  @InternalUseOnly
  public double getDataStart () {
    return dataStart;
  }

  /**
   * @return the extSize
   */
  @InternalUseOnly
  public int getExtSize () {
    return extSize;
  }

  /**
   * @return the extStart
   */
  @InternalUseOnly
  public int getExtStart () {
    return extStart;
  }

  /**
   * @return the version
   */
  @InternalUseOnly
  public String getVersion () {
    return version;
  }

  /**
   * @return the pipeId
   */
  @InternalUseOnly
  public int getPipeId () {
    return pipeId;
  }

  /**
   * @return the detached
   */
  @InternalUseOnly
  public int getDetached () {
    return detached;
  }

  /**
   * @return the protectedFromOverwrite
   */
  @InternalUseOnly
  public int getProtectedFromOverwrite () {
    return protectedFromOverwrite;
  }

  /**
   * @return the timeCode
   */
  @InternalUseOnly
  public double getTimeCode () {
    return timeCode;
  }

  /**
   * @return the keyLength
   */
  @InternalUseOnly
  public int getKeyLength () {
    return keyLength;
  }

  /**
   * @return the inlet
   */
  @InternalUseOnly
  public short getInlet () {
    return inlet;
  }

  /**
   * @return the outlet
   */
  @InternalUseOnly
  public short getOutlet () {
    return outlet;
  }

  /**
   * @return the pipeLoc
   */
  @InternalUseOnly
  public int getPipeLoc () {
    return pipeLoc;
  }

  /**
   * @return the pipeSize
   */
  @InternalUseOnly
  public int getPipeSize () {
    return pipeSize;
  }

  /**
   * @return the inByte
   */
  @InternalUseOnly
  public double getInByte () {
    return inByte;
  }

  /**
   * @return the outByte
   */
  @InternalUseOnly
  public double getOutByte () {
    return outByte;
  }

  /**
   * @return the hbuf
   */
  @InternalUseOnly
  public byte[] getHbuf () {
    return hbuf;
  }

  /**
   * @return the res
   */
  @InternalUseOnly
  public IOResource getRes () {
    return res;
  }

  /**
   * @return the ableToParseHeader
   */
  @InternalUseOnly
  public boolean isAbleToParseHeader () {
    return ableToParseHeader;
  }

}
