package nxm.sys.lib;

import static nxm.sys.lib.IOResource.UNKNOWN;

import java.net.URL;

import nxm.sys.inc.IOResourceFactory;
import nxm.sys.inc.InternalUseOnly;

/** A new resource that will used the current Midas Context's ClassLoader to access the files
 *  in the classpath. <br>
 *  <br>
 *  NOTE: This resource type will generally provide read-only access consistent with using the
 *  Java ClassLoader to access the files. In some cases the file will be read/write due to the
 *  ClassLoader finding the file on a location within the classpath that is writable. However,
 *  this will NOT find any output-only files.
 *  @author Juan Velasquez
 *  @author J. Czechowski IV
 *  @since  NeXtMidas 3.7.2
 */
public class ClassPathResource implements IOResourceFactory {
  static final ClassPathResource INSTANCE = new ClassPathResource();

  @Override
  public IOResource getResourceFor (Midas M, int type, String urlstr, int dir, Table quals) {
    if (!urlstr.startsWith("classpath:")) {
      throw new MidasException("Invalid URL for ClassPathResource: "+urlstr);
    }
    int    i        = urlstr.indexOf('!');             // see if it is in the form "classpath:foo.tar!/myFile.txt"
    String filePath = (i < 0)? urlstr.substring(10)    // strip off the "classpath:"
                             : urlstr.substring(10,i); // strip off the "classpath:" and "!..."
    String suffix   = (i < 0)? "" : urlstr.substring(i);
    URL    fileURL  = getResource(filePath);

    if (fileURL != null) {
      // Use the Midas CoreIO to find the resource needed for the URL found on the classpath
      return IOResource.getInstance(M, UNKNOWN, fileURL+suffix, dir, quals);
    }
    return null;
  }

  /** Load a given resource. This method will try to load the resource using the following methods
   *  (in order):
   *  <pre>
   *    Thread.currentThread().getContextClassLoader()
   *    Shell.class.getClassLoader()
   *    ClassLoader.getSystemClassLoader().getClassLoader()
   *  </pre>
   *  @param pathName The name of the resource to load
   */
  @InternalUseOnly
  private URL getResource (String pathName) {
    URL url;

    // Normalize the path (replace '\' with '/' and replace '//' with '/')
    pathName = pathName.replace('\\', '/').replaceAll("/+", "/");
    if (pathName.startsWith("/")) {
      pathName = pathName.substring(1); // take off leading '/'
    }

    // Get the URL from the class loader
    url = Thread.currentThread().getContextClassLoader().getResource(pathName);
    if (url != null) return url;

    url = Shell.class.getClassLoader().getResource(pathName);
    if (url != null) return url;

    url = ClassLoader.getSystemClassLoader().getResource(pathName);
    if (url != null) return url;

    return null;
  }
}